/*
   Coolpix RAW to Coolpix NEF converter
   by paul69, v0.07, 2003-01-31, <e2500@narod.ru>

   Main function

   Supported camera (tested)
      E2500
      E4500
      E995
      E990
      E950
      E900 - Thanks to Sakura Shan
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#pragma hdrstop
#include "tiff.h"
#include "file.h"
#include "raw2nef.h"

#define program "raw2nef v0.07, by paul69 <e2500@narod.ru> [compiled: " __DATE__ " " __TIME__ "]\n"

bool raw_to_nef(int n, char* a[]);

int main(int argc, char* argv[] )
{
   return raw_to_nef( argc, argv );
}

static void usage();
static char ofn_buf[_MAX_PATH];
static char efn_buf[_MAX_PATH];

static
TCCDParam default_pars[] = {
   {1600+16, 1200+4, 0x2CE000, "YCGM", 2424, 12, 2500, {{ 0, (1200+4)/2 }, {0x2CE000/2, (1200+4)/2}}  }, // E2500
   {2272+16, 1704+4, 0x598000, "GMYC", 3432, 12, 4500, {{ 0, (1704+4)/2 }, {0x598000/2, (1704+4)/2}}  }, // E4500
   {2048+16, 1536+4, 0x48D000, "MGCY", 3096, 12,  995, {{ 0, (1536+4)/2 }, {0x48D000/2, (1536+4)/2}}  }, // E995
   {2048+16, 1536+4, 0x48D000, "GMYC", 3096, 12,  990, {{ 0, (1536+4)/2 }, {0x48D000/2, (1536+4)/2}}  }, // E990
   {1600+16, 1200+4, 0x25A000, "YCGM", 2048, 10,  950, {{ 0, (1200+4)/2 }, {0x25A000/2, (1200+4)/2}}  }, // E950
   {1280+16,  960+4,0x182000+4,"CYMG", 1632, 10,  900, {{ 0,  (960+4)/2 }, {0x182000/2, ( 960+4)/2}}  }, // E900
};

const int n_default_pars = sizeof(default_pars) / sizeof(default_pars[0]);

char* CreateExifFileName(const char* FilePath, int NameOffset, int ExtOffset)
{
   if( NameOffset + 8 != ExtOffset )
      return NULL;

   const char* FileName = FilePath+NameOffset;
   if( (FileName[0] & ~0x20) != 'D' ) return NULL;
   if( (FileName[1] & ~0x20) != 'S' ) return NULL;
   if( (FileName[2] & ~0x20) != 'C' ) return NULL;

   const char* NameTemplate = "DSCN%04d.JPG";
   const char* FileNum = FileName+4;

   if( (FileName[3] & ~0x20) != 'N' )
   {
      if( FileName[3] < '0' || FileName[3] > '9' ) return NULL;

      NameTemplate = "DSC%05d.JPG";
      --FileNum;
   }
   
   int n = atoi( FileNum );
   if( n <= 0 ) return NULL;

   char* efn = efn_buf;
   memcpy(efn, FilePath, NameOffset);
   sprintf(efn + NameOffset, NameTemplate, n+1);
   return efn_buf;
}

bool ProcessFile(const char* ifn, char* ofn, char* efn, unsigned camera, bool verbose, bool copy_exif, bool endian)
{
   int i;

   if( verbose )
      printf("input: %s\n", ifn);

   if( !ofn || !efn )
   {
      const char* inpFileName = ifn;

      const char* ifn_slash = strrchr(inpFileName, '/');
      if( ifn_slash )
         inpFileName = ifn_slash+1;

      const char* ifn_bslash = strrchr(inpFileName, '\\');
      if( ifn_bslash )
         inpFileName = ifn_bslash+1;

      const char* inpFileExt = inpFileExt = strrchr(inpFileName, '.');

      if( !inpFileExt )
         inpFileExt = ifn + strlen(ifn);

      if( !ofn )
      {
         ofn = ofn_buf;
         int len = inpFileExt - ifn;
         memcpy(ofn, ifn, len);
         strcpy(ofn+len, ".NEF");
      }

      if( !efn && copy_exif )
      {
         efn = CreateExifFileName(ifn, inpFileName-ifn, inpFileExt-ifn);
         if( !efn && verbose )
            printf("warning: cannot get exif filename\n");
      }
   }

   TFile in = fopen( ifn, "rb");
   if( !in ) { printf("error: \"%s\" - cannot open\n", ifn); return false; }

   fseek( in, 0, SEEK_END );
   unsigned file_size = ftell( in );

   NEF_IFD Nef_IFD;

   if( efn )
   {
      if( verbose )
         printf("read exif info from \"%s\"\n", efn);

      TFile exif = fopen( efn, "rb");

      if( !exif && verbose )
         { printf("warning: \"%s\" - cannot open exif file\n", efn); }
      else
      if( !Nef_IFD.read_exif( exif ) && verbose )
         { printf("warning: \"%s\" - cannot read exif file\n", efn); }
   }

   IFDir::Tag* Model = Nef_IFD.IFD0.get_tag( TIFF::Model );
   if( Model && ((const char*)Model->value)[0] == 'E' )
   {
      unsigned exif_camera = atoi( (const char*)Model->value+1 );
      if( exif_camera != 0 )
      {
         if( camera == 0 )
            camera = exif_camera;
         else
         if( camera != exif_camera )
         { printf("warning: exif model E%u and option model E%u is not same\n", exif_camera, camera); }
      }
   }
   
   for(i=0; i<n_default_pars; ++i)
   {
      if( camera == 0 )
      {
         if( file_size == default_pars[i].file_size )
            break;
      }
      else
      {
         if( camera == default_pars[i].camera )
            break;
      }
   }

   TCCDParam ccd_pars;

   if( i < n_default_pars )
   {
      if( verbose )
         printf("camera model is E%u\n", default_pars[i].camera );

      ccd_pars = default_pars[i];
   }
   else
   {
      printf("error: unsupported format\n");
      return false;
   }

   TBuff ccd(ccd_pars.raw_height*ccd_pars.raw_width);

   if( !ccd )
   {
      printf("error: memory allocation fault (%u)\n", ccd_pars.raw_height*ccd_pars.raw_width*sizeof(color_t));
      return false;
   }

   if( verbose )
      printf("read raw data from \"%s\"\n", ifn);

   if( !read_raw( in, ccd, ccd_pars) )
   {
      printf("error: can not read raw data\n");
      return false;
   }

   in.close();

   if( verbose )
      printf("write picture to: \"%s\"\n", ofn);

   TFile out = fopen( ofn, "wb");
   if( !out ) { printf("error: \"%s\" - can not open\n", ofn); return false; }

   if( !Nef_IFD.write_nef(out, endian, ccd, ccd_pars) )
      { printf("error: can not write NEF\n"); return false; }

   if( verbose )
      printf("OK\n");

   return true;
}

bool raw_to_nef(int n, char* a[])
{
   // endian: true for Intel, false for Motorolla
   bool endian = true;
   bool copy_exif = true;
   char* ifn = NULL;
   char* ofn = NULL;
   char* efn = NULL;
   bool verbose = false;
   bool batch_mode = true;
   unsigned camera = 0;
   
   if( n <= 1 )
   {
      usage();
      return false;
   }

   int i;
   for(i=1; i<n; ++i)
   {
      if( a[i][0] == '-' )
      {
         switch( a[i][1] )
         {
         case 'v': verbose = true; break;

         case 'e':
            batch_mode = false;
            if( i+1 < n && a[i+1][0] != '-' )
               efn = a[++i];
            else
               goto missing;
            break;

         case 'o':
            batch_mode = false;
            if( i+1 < n && a[i+1][0] != '-' )
               ofn = a[++i];
            else
               goto missing;
            break;

         case '?': usage(); return false;

         case 'a': camera = 0; break;

         default:
            if( a[i][1] >= '1' && a[i][1] <= '9' )
               camera = atoi( a[i]+1 );
            else
               printf("warning: unknown switch: %s\n", a[i]);
            break;

         invalid:
            printf("warning: invalid switch: %s\n", a[i]);
            break;

         missing:
            printf("warning: missing argument for %s\n", a[i]);
            break;
         }
      }
      else
      {
         if( !ifn )
            ifn = a[i];
      }
   }

   if( verbose )
      printf( program );

   if( !ifn )
   {
      printf("error: input file is missing\n\n");
      return false;
   }

   if( batch_mode )
   {
      for(i=1; i<n; ++i)
      {
         if( a[i][0] == '-' )
         {
            switch( a[i][1] )
            {
            case 'o':
            case 'e':
               if( i+1 < n && a[i+1][0] != '-' )
                  ++i;
               break;
            }
         }
         else
         {
            if( a[i][0] )
            {
               if( !ProcessFile(a[i], NULL, NULL, camera, verbose, copy_exif, endian) )
                  printf("cannot process \"%s\"\n\n", a[i]);
            }
         }
      }
   }
   else
   {
      if( !ProcessFile(ifn, ofn, efn, camera, verbose, copy_exif, endian) )
         printf("cannot process \"%s\"\n", ifn);
   }
   return true;
}

void usage()
{
   printf("Coolpix RAW data to E5700 NEF file converter\n");
   printf( program );

   printf("\nargs: [options] input_file1 [input_file2...]\n");

   printf("\noptions: camera\n");
   printf("   -auto - autodetect (default)\n");
   printf("   -2500 - Coolpix 2500\n");
   printf("   -4500 - Coolpix 4500\n");
   printf("   -900 - Coolpix 900\n");
   printf("   -950 - Coolpix 950\n");
   printf("   -990 - Coolpix 990 (auto is wrong for 990!)\n");
   printf("   -995 - Coolpix 995\n");

   printf("\noptions: processing\n");
   printf("   -v - verbose\n");
   printf("   -e <filename> - exif jpeg file name (disable batch mode)\n");
   printf("   -o <filename> - nef output file name (disable batch mode)\n");

   printf("\nSupported cameras: E2500, E4500, E995, E990, E950, E900\n");
}

